# encoding: utf-8

"""
Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import json
import numpy
import pyproj
import logging
try:
    from osgeo import osr
except ImportError:
    import osr

class LevelFilter(logging.Filter):
    def __init__(self, levels):
        self.levels = levels

    def filter(self, record):
        return record.levelno in self.levels

class NumpyAwareJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, numpy.ndarray) and obj.ndim == 1:
            return obj.tolist()
        elif isinstance(obj, numpy.float32):
            return obj.item()
        elif isinstance(obj, numpy.int32):
            return obj.item()
        return json.JSONEncoder.default(self, obj)


class CoordinatesShapeNotSupported(ValueError):
    """Error raised when the coordinates to transform are not points in either
    (x, y) or (x, y, z) format."""
    pass


def transform_coordinates(input_srs, output_srs, coordinates):
    """"""
    input_proj = pyproj.Proj(input_srs.ExportToProj4())
    output_proj = pyproj.Proj(output_srs.ExportToProj4())
    _, dims = numpy.array(coordinates).shape
    tcoordinates = None
    if 2 == dims:
        x, y = list(zip(*coordinates))
        tx, ty = pyproj.transform(input_proj, output_proj, x, y)
        tcoordinates = list(zip(tx, ty))
    elif 3 == dims:
        x, y, z = list(zip(*coordinates))
        tx, ty, tz = pyproj.transform(input_proj, output_proj, x, y, z)
        tcoordinates = list(zip(tx, ty, tz))
    else:
        raise CoordinatesShapeNotSupported(dims)

    return tcoordinates


def epsg2srs(epsg, plusover=False):
    """"""
    srs = osr.SpatialReference()
    if plusover is True:
        tmp_srs = osr.SpatialReference()
        tmp_srs.ImportFromEPSG(epsg)
        srs.ImportFromProj4(tmp_srs.ExportToProj4() + ' +over')
    else:
        srs.ImportFromEPSG(epsg)

    if hasattr(srs, 'SetAxisMappingStrategy'):
        srs.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
    return srs


def epsg2wkt(epsg):
    """Translate EPSG code into a WKT"""
    srs = epsg2srs(epsg)
    return srs.ExportToWkt()


def epsg2proj4(epsg, plusover=False):
    """"""
    srs = epsg2srs(epsg)
    proj4 = srs.ExportToProj4()
    if plusover is True:
        proj4 += ' +over'
    return proj4


def wkt2epsg(projection):
    """"""
    srs = osr.SpatialReference()
    srs.ImportFromWkt(projection)
    srs.AutoIdentifyEPSG()
    epsg = int(srs.GetAuthorityCode(None))
    return epsg
